@@ -2,7 +2,8 @@ mod capabilities;
 pub mod citation;
 mod client;
 pub mod component_db;
-pub mod diagnostics;
+mod debouncer;
+mod diagnostics;
 mod dispatch;
 pub mod distro;
 mod document;
