@@ -1,11 +1,13 @@
 #[cfg(feature = "completion")]
 mod completion;
 mod definition;
+mod did_change_configuration;
 mod document_highlight;
 mod document_link;
 mod document_symbol;
 mod folding_range;
 mod formatting;
 mod hover;
+mod publish_diagnostics;
 mod references;
 mod rename;
